#!/usr/bin/python
import operator
from pprint import pprint

class MachineStop(Exception):
    pass

class Fault(Exception):
    pass

class Debug(Exception):
    pass

def shorten(v, l=30):
    v = repr(v)
    if len(v) > l:
        v = v[:l].rstrip() + '...'
    return v

class Frame(object):
    def __init__(self, parent, data):
        self.parent = parent
        self.data = data

    def _find_frame(self, depth):
        f = self
        for i in range(depth):
            f = f.parent
        return f

    def get(self, depth, idx):
        #print 'curr', self
        #print 'found', self._find_frame(depth)
        return self._find_frame(depth).data[idx]

    def set(self, depth, idx, v):
        self._find_frame(depth).data[idx] = v

    def __repr__(self):
        return '[' + ', '.join(shorten(v) for v in self.data) + ']'

class Closure(object):
    def __init__(self, fp, env):
        self.fp = fp
        self.env = env

    def __repr__(self):
        return 'Closure(%s, %s)' % (self.fp, self.env)

def check_int(*args):
    for i in args:
        if not isinstance(i, int):
            raise Exception("Integer expected, got '%s'" % i)

class GccMachine(object):
    def __init__(self, program, dstack, cstack, env):
        self.code = self._parse_program(program)
        self.c = 0 # program counter
        self.s = dstack
        self.d = cstack
        self.e = env

    def _parse_program(self, program):
        code = []
        for l in program.splitlines():
            l = l.strip()
            parts = l.split(';')[0].strip().split()
            if not parts:
                continue
            op = parts[0]
            args = map(int, parts[1:])
            code.append((op, args, l))
        return code

    def step(self):
        def stack_binop(f, is_int=True):
            y, x = self.s.pop(), self.s.pop()
            if is_int:
                check_int(x, y)
            self.s.append(f(x, y))
            self.c +=1

        op, a, codeline = self.code[self.c]
        if op == 'LDC':
            self.s.append(a[0])
            self.c +=1
        elif op == 'LD':
            self.s.append(self.e.get(a[0], a[1]))
            self.c +=1
        elif op == 'ADD': stack_binop(operator.add)
        elif op == 'SUB': stack_binop(operator.sub)
        elif op == 'MUL': stack_binop(operator.mul)
        elif op == 'DIV': stack_binop(operator.div)
        elif op == 'CEQ': stack_binop(lambda x, y: int(x == y))
        elif op == 'CGT': stack_binop(lambda x, y: int(x > y))
        elif op == 'CGTE': stack_binop(lambda x, y: int(x >= y))
        elif op == 'ATOM':
            x = self.s.pop()
            self.s.append(int(isinstance(x, int)))
            self.c += 1
        elif op == 'CONS': stack_binop(lambda x, y: (x, y), False)
        elif op == 'CAR':
            x = self.s.pop()
            self.s.append(x[0])
            self.c += 1
        elif op == 'CDR':
            x = self.s.pop()
            self.s.append(x[1])
            self.c += 1
        elif op == 'SEL':
            x = self.s.pop()
            check_int(x)
            self.d.append((self.c + 1, 'JOIN'))
            self.c = a[1] if x == 0 else a[0]
        elif op == 'JOIN':
            x, tag = self.d.pop()
            if tag != 'JOIN':
                raise Exception('Tag mismatch, got: %s' % tag)
            self.c = x
        elif op == 'LDF':
            check_int(a[0])
            self.s.append(Closure(a[0], self.e))
            self.c +=1
        elif op == 'AP':
            x = self.s.pop()
            if not isinstance(x, Closure):
                raise Exception("Expected Closure, got '%s'" % x)
            env_data = [self.s.pop() for i in xrange(a[0])][::-1]
            env = Frame(x.env, env_data)
            self.d.append((self.e, 'FRAME'))
            self.d.append((self.c + 1, 'RET'))
            self.e = env
            self.c = x.fp
        elif op == 'RTN':
            c, tag = self.d.pop()
            if tag == 'STOP':
                raise MachineStop('stop')
            if tag != 'RET':
                raise Fault("CONTROL MISMATCH: Expected RET, got '%s'" % tag)
            e, tag = self.d.pop()
            if tag != 'FRAME':
                raise Fault("Expected FRAME, got '%s'" % tag)
            self.e = e
            self.c = c
        elif op == 'DUM': raise NotImplemented('DUM')
        elif op == 'RAP': raise NotImplemented('RAP')
        elif op == 'STOP': raise MachineStop('stop')
        elif op == 'TSEL':
            x = self.s.pop()
            check_int(x)
            self.c = a[1] if x == 0 else a[0]
        elif op == 'TAP': raise NotImplemented('TAP')
        elif op == 'TRAP': raise NotImplemented('TRAP')
        elif op == 'ST':
            # fix if DUM is implemented
            x = self.s.pop()
            self.e.set(a[0], a[1], x)
            self.c += 1
        elif op == 'DBUG':
            x = self.s.pop()
            self.c += 1
            print 'DEBUG', x
            raise Debug(str(x))
        else: raise Fault('Unknown command')

    def run(self, debug=0):
        cnt = 0
        while 1:
            cnt += 1
            try:
                try:
                    self.step()
                except Debug:
                    if debug > 0:
                        debug = 3

                if debug > 1:
                    print '-'*70, cnt
                    self.print_state()
                if debug > 2:
                    print 'Stopping on DBUG interrupt'
                    print '<Enter> to step, number 0..3 to set debug level for this step'
                    inp = raw_input()
                    inp = inp.strip()
                    if inp:
                        debug = int(inp)
            except MachineStop:
                return cnt

    def print_state(self):
        l0 = max(self.c - 2, 0)
        l1 = min(self.c + 4, len(self.code) - 1)
        for i in range(l0, l1+1):
            marker = '->' if i == self.c else '  '
            print '  %s%4d  %s' % (marker, i, self.code[i][2])

        print 'frame:'
        for i, d in enumerate(self.e.data):
            print '    ', shorten(d, 72)

        print 'stack: (%d)' % len(self.s)
        for i in self.s[::-1]:
            print '    ', shorten(i, 72)

        print 'control: (%d)' % len(self.s)
        for i in self.d[::-1]:
            print '    ', i


def to_gcc_tuple(arr):
    tp = (arr[-2], arr[-1])
    for i in arr[-3::-1]:
        tp = (i, tp)
    return tp


def to_gcc_list(arr):
    tp = (arr[-1], 0)
    for i in arr[-2::-1]:
        tp = (i, tp)
    return tp


if __name__ == '__main__':
    import sys

    prog = file(sys.argv[1]).read()

    f = Frame(None, [
        to_gcc_list([1,2,3,4,5,6,7]),
        2,
        55,
        "tmp"
    ])

    m = GccMachine(prog, [], [(None, 'STOP')], f)
    m.run(debug=4)
    #while 1:
        #m.step()
        #m.print_state()
        #raw_input()

